// Copyright (C) 2008
// by all contributors.

// This file is part of SCNSL.

// SCNSL is free software:
// you can redistribute it and/or modify it under the terms of the
// GNU Lesser General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your option)
// any later version.

// SCNSL is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.
// See the GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with SCNSL,
// in a file named LICENSE.txt.
// If not, see <http://www.gnu.org/licenses/>.



/// @file
/// A base class for basic RTL tasks.

#include <string>
#include <sstream>

#include "scnsl/rtl/RtlTask_base_if_t.hh"


using Scnsl::Rtl::RtlTask_base_if_t;

// ////////////////////////////////////////////////////////////////
// Destructor.
// ////////////////////////////////////////////////////////////////

RtlTask_base_if_t::~RtlTask_base_if_t()
{
    // Calling destructors for all ports.
    // NOTE: reverse order not truly required here,
    // while it is a must in the STL.
    for ( register unsigned int i = _PROXIES_NUMBER; i != 0; )
    {
        --i;

        // Ports for inputs:
        carrier[ i ].sc_core::sc_in< carrier_t >::~sc_in();
        inputPacketSize[ i ].sc_core::sc_in< size_t >::~sc_in();
        inputPacketLabel[ i ].sc_core::sc_in< label_t >::~sc_in();
        newInputPacket[ i ].sc_core::sc_in< counter_t >::~sc_in();

        // Ports for outputs:
        outputPacketSize[ i ].sc_core::sc_out< size_t >::~sc_out();
        newOutputPacket[ i ].sc_core::sc_out< counter_t >::~sc_out();
        outputPacketLabel[ i ].sc_core::sc_out< label_t >::~sc_out();
        packetSendCompleted[ i ].sc_core::sc_in< counter_t >::~sc_in();
    }

    // Deallocating the memory.

    // Ports for inputs:
    ::operator delete[] ( carrier );
    ::operator delete[] ( inputPacketSize );
    ::operator delete[] ( inputPacketLabel );
    ::operator delete[] ( newInputPacket );

    // Ports for outputs:
    ::operator delete[] ( outputPacketSize );
    ::operator delete[] ( newOutputPacket );
	::operator delete[] ( outputPacketLabel );
    ::operator delete[] ( packetSendCompleted );
}


// ////////////////////////////////////////////////////////////////
// Interface methods.
// ////////////////////////////////////////////////////////////////

RtlTask_base_if_t::size_t RtlTask_base_if_t::getProxiesNumber()
    const 
{
    SCNSL_TRACE_DBG( 5, "<> getProxiesNumber()." );
    return _PROXIES_NUMBER;
}

RtlTask_base_if_t::size_t RtlTask_base_if_t::getBoundedProxiesNumber()
    const 
{
    SCNSL_TRACE_DBG( 5, "<> getBoundedProxiesNumber()." );
    return _bounded_proxies;
}



// ////////////////////////////////////////////////////////////////
// Protected methods.
// ////////////////////////////////////////////////////////////////


RtlTask_base_if_t::RtlTask_base_if_t( const sc_core::sc_module_name modulename,
                                      const task_id_t id,
                                      Scnsl::Core::Node_t * n,
                                      const size_t proxies )
    :
    // Parent:
    Task_if_t( modulename, id, n ),
    // Ports for inputs:
    carrier( NULL ),
    inputPacketSize( NULL ),
	inputPacketLabel( NULL ),
    newInputPacket( NULL ),
    // Ports for outputs:
    outputPacketSize( NULL ),
    newOutputPacket( NULL ),
	outputPacketLabel( NULL ),
    packetSendCompleted( NULL ),
    // Fields:
    _PROXIES_NUMBER( proxies ),
    _bounded_proxies( 0 )
{
    if ( _PROXIES_NUMBER == 0 )
    {
        throw std::invalid_argument( "Task with 0 proxies associated." );
    }

    // Initializing the ports.
    // Done in this way to permit a good naming.

    // Allocating the memory.

    // Ports for inputs:
    carrier = static_cast< sc_core::sc_in< carrier_t > * >(
        ::operator new[] ( _PROXIES_NUMBER * sizeof(sc_core::sc_in< carrier_t >) ));
    inputPacketSize = static_cast< sc_core::sc_in< size_t > * >(
        ::operator new[] ( _PROXIES_NUMBER * sizeof(sc_core::sc_in< size_t >) ));
    inputPacketLabel = static_cast< sc_core::sc_in< label_t > * >(
        ::operator new[] ( _PROXIES_NUMBER * sizeof(sc_core::sc_in< label_t >) ));
    newInputPacket = static_cast< sc_core::sc_in< counter_t > * >(
        ::operator new[] ( _PROXIES_NUMBER * sizeof(sc_core::sc_in< counter_t >) ));

    // Ports for outputs:
    outputPacketSize = static_cast< sc_core::sc_out< size_t > * >(
        ::operator new[] ( _PROXIES_NUMBER * sizeof(sc_core::sc_out< size_t >) ));
    newOutputPacket = static_cast< sc_core::sc_out< counter_t > * >(
        ::operator new[] ( _PROXIES_NUMBER * sizeof(sc_core::sc_out< counter_t >) ));
    outputPacketLabel = static_cast< sc_core::sc_out< label_t > * >(
        ::operator new[] ( _PROXIES_NUMBER * sizeof(sc_core::sc_out< label_t >) ));
    packetSendCompleted = static_cast< sc_core::sc_in< counter_t > * >(
        ::operator new[] ( _PROXIES_NUMBER * sizeof(sc_core::sc_in< counter_t >) ));

    // Calling ports constructors.
    for ( register unsigned int i = 0; i < _PROXIES_NUMBER; ++i )
    {
        std::string s;
        std::stringstream ss;
        ss << i;
        ss >> s;

        // Ports for inputs:
        new (carrier + i) sc_core::sc_in< carrier_t >(("carrier" + s).c_str());
        new (inputPacketSize + i) sc_core::sc_in< unsigned int >(("inputPacketSize" + s).c_str());
        new (inputPacketLabel + i) sc_core::sc_in< label_t >(("inputPacketLabel" + s).c_str());
        new (newInputPacket + i) sc_core::sc_in< counter_t >(("newInputPacket" + s).c_str());

        // Ports for outputs:
        new (outputPacketSize + i) sc_core::sc_out< size_t >(("outputPacketSize" + s).c_str());
        new (newOutputPacket + i) sc_core::sc_out< counter_t >(("newOutputPacket" + s).c_str());
        new (outputPacketLabel + i) sc_core::sc_out< label_t >(("outputPacketLabel" + s).c_str());
        new (packetSendCompleted + i) sc_core::sc_in< counter_t >(("packetSendCompleted" + s).c_str());
    }
}
