#include <cmath>
#include <limits>

#include "scnsl/utils/DirectionalEnvironment_t.hh"

using Scnsl::Utils::DirectionalEnvironment_t;
using namespace Scnsl::Core;

const double DirectionalEnvironment_t::omnidirectional=360;

DirectionalEnvironment_t::DirectionalEnvironment_t(
        const double alpha, const Scnsl::Core::propagation_t propagation):
    _ALPHA_EXPONENT(alpha),
    _PROPAGATION(propagation)
{
    // ntd
}

DirectionalEnvironment_t::~DirectionalEnvironment_t()
{
    // Nothing to do.
}

void DirectionalEnvironment_t::createInstance(const double alpha,
                                              const propagation_t propagation)
{
    static DirectionalEnvironment_t p(alpha, propagation);
    DirectionalEnvironment_t::_env = &p;
}

double DirectionalEnvironment_t::_getDistance(const node_properties_t & sp, 
                                              const node_properties_t & rp)
{

    const position_t dx = rp.x - sp.x;
    const position_t dy = rp.y - sp.y;
    const position_t dz = rp.z - sp.z;
    const double d2 = (dx * dx) + (dy * dy) + (dz * dz);

    return sqrt(d2);
}

bool DirectionalEnvironment_t::_doubleEqual(const double a, const double b)
{
    return fabs(a - b) < std::numeric_limits<double>::epsilon();
}

double DirectionalEnvironment_t::_getDirectionAngle(const node_properties_t & sp, 
                                                    const unsigned j, const node_properties_t & rp)
{
    const double norm = sqrt( sp.lobes[j].direction[0] * sp.lobes[j].direction[0]
            + sp.lobes[j].direction[1] * sp.lobes[j].direction[1]
            + sp.lobes[j].direction[2] * sp.lobes[j].direction[2]);

    const double distance = _getDistance(sp, rp);

    const position_t sr_x = (rp.x - sp.x) / distance;
    const position_t sr_y = (rp.y - sp.y) / distance;
    const position_t sr_z = (rp.z - sp.z) / distance;

    const position_t norm_x = sp.lobes[j].direction[0] / norm;
    const position_t norm_y = sp.lobes[j].direction[1] / norm;
    const position_t norm_z = sp.lobes[j].direction[2] / norm;
    const double cos = (sr_x * norm_x) + (sr_y * norm_y) + (sr_z * norm_z);

    return acos(cos) * 180 / M_PI;
}

double DirectionalEnvironment_t::getReceiverPower(const node_properties_t & sp,
                                                  const node_properties_t & rp)
{
    double power = 0;
    double distance = 0;
    distance = _getDistance(sp, rp);

    // Omnidirectional source NODE and Omnidirectional receiver NODE
    if(sp.lobes.empty() && rp.lobes.empty())
    {
        power = (sp.transmission_power / pow( distance, _ALPHA_EXPONENT ));
        return power;
    }
    
    // Omnidirectional source NODE and Directional receiver NODE
    if(sp.lobes.empty() && !(rp.lobes.empty()))
    {
        for(unsigned j=0;j < rp.lobes.size();++j)
        {
            // Skip inactive lobes
            if(_doubleEqual(rp.lobes[j].gain, 0))
                continue;

            // Omnidirectional receiver LOBE
            if(_doubleEqual(rp.lobes[j].beamwidth, omnidirectional))
            {
                power += ((sp.transmission_power * rp.lobes[j].gain)
                          / pow( distance, _ALPHA_EXPONENT ));
            }

            // Directional receiver LOBE
            else
            {
                double source_angle = _getDirectionAngle(rp, j, sp);
                if((rp.lobes[j].beamwidth / 2) > source_angle)
                {
                    power += ((sp.transmission_power * rp.lobes[j].gain)
                              / pow( distance, _ALPHA_EXPONENT ));
                }
            }
        }
    }

    // Directional source NODE and Omnidirectional receiver NODE
    if( !(sp.lobes.empty()) && rp.lobes.empty())
    {
        for(unsigned i=0; i<sp.lobes.size();++i)
        {
            // Skip inactive lobes
            if(_doubleEqual(sp.lobes[i].gain, 0))
                continue;

            // Omnidirectional source LOBE
            if(_doubleEqual(sp.lobes[i].beamwidth, omnidirectional))
            {
                power += ((sp.transmission_power * sp.lobes[i].gain)
                          / pow( distance, _ALPHA_EXPONENT ));
            }

            // Directional source LOBE
            else
            {
                double receiver_angle = _getDirectionAngle(sp, i, rp);
                if((sp.lobes[i].beamwidth / 2) > receiver_angle)
                {
                    power += ((sp.transmission_power * sp.lobes[i].gain)
                              / pow( distance, _ALPHA_EXPONENT ));
                }
            }
        }
    }

    // Directional source NODE and Directional receiver NODE
    if( !(sp.lobes.empty()) && !(rp.lobes.empty()) )
    {
        for(unsigned i=0; i<sp.lobes.size();++i)
        {
            // Skip inactive lobes
            if(_doubleEqual(sp.lobes[i].gain, 0)) continue;

            for(unsigned j=0;j<rp.lobes.size();++j)
            {
                // Skip inactive lobes
                if(_doubleEqual(rp.lobes[j].gain, 0)) continue;

                // Omnidirectional source LOBE and Omnidirectional Receiver LOBE
                if(  _doubleEqual(sp.lobes[i].beamwidth, omnidirectional) &&
                     _doubleEqual(rp.lobes[j].beamwidth, omnidirectional))
                {
                    power += ((sp.transmission_power * rp.lobes[j].gain
                               * sp.lobes[i].gain) / pow( distance, _ALPHA_EXPONENT ));
                }

                // Omnidirectional source LOBE and Directional receiver LOBE
                if(  _doubleEqual(sp.lobes[i].beamwidth, omnidirectional) &&
                     ! _doubleEqual(rp.lobes[j].beamwidth, omnidirectional))
                {
                    double source_angle = _getDirectionAngle(rp, j, sp);
                    if((rp.lobes[j].beamwidth / 2) > source_angle)
                    {
                        power += ((sp.transmission_power * rp.lobes[j].gain
                                   * sp.lobes[i].gain) / pow( distance, _ALPHA_EXPONENT ));
                    }
                }


                // Directional source LOBE and Omnidirectional receiver LOBE
                if(! _doubleEqual(sp.lobes[i].beamwidth, omnidirectional) &&
                   _doubleEqual(rp.lobes[j].beamwidth, omnidirectional))
                {
                    double receiver_angle = _getDirectionAngle(sp, i, rp);
                    if((sp.lobes[i].beamwidth / 2) > receiver_angle)
                    {
                        power += ((sp.transmission_power * rp.lobes[j].gain
                                   * sp.lobes[i].gain) / pow( distance, _ALPHA_EXPONENT ));
                    }
                }

                // Directional Receiver LOBE and Directional receiver LOBE
                if(! _doubleEqual(sp.lobes[i].beamwidth, omnidirectional) &&
                   ! _doubleEqual(rp.lobes[j].beamwidth, omnidirectional))
                {
                    double receiver_angle = _getDirectionAngle(sp, i, rp);
                    if((sp.lobes[i].beamwidth / 2) < receiver_angle) continue;

                    double source_angle = _getDirectionAngle(rp, j, sp);
                    if((rp.lobes[j].beamwidth / 2) > source_angle)
                    {
                        power += ((sp.transmission_power * rp.lobes[j].gain
                                   * sp.lobes[i].gain) / pow( distance, _ALPHA_EXPONENT ));
                    }
                }
            }
        }
    }

    return power;
}

double DirectionalEnvironment_t::getMaxDelay( const node_properties_t & sp,
                                              const node_properties_t & rp)
{
    return pow(sp.transmission_power / rp.receiving_threshold,
               1 / _ALPHA_EXPONENT) / _PROPAGATION;

}

double DirectionalEnvironment_t::getBitErrorRate(const node_properties_t & /*sp*/,
                                                 const node_properties_t & /*rp*/)
{
    return 0.0;
}

Scnsl::Core::propagation_t DirectionalEnvironment_t::getPropagation()
{
    return _PROPAGATION;
}
