// Copyright (C) 2008
// by all contributors.

// This file is part of SCNSL.

// SCNSL is free software:
// you can redistribute it and/or modify it under the terms of the
// GNU Lesser General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your option)
// any later version.

// SCNSL is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.
// See the GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with SCNSL,
// in a file named LICENSE.txt.
// If not, see <http://www.gnu.org/licenses/>.


// Description
//
// This program shows how to create automatically a wireless MESH topology for
// temperature monitoring in a building.
// The environment taken as an example is a building of 5 floors, 3 corridors
// for each floor and 10 rooms for each corridor.
// The rooms are represented by the z-coordinate, corridors by the y-coordinate
// and floors by the x-coordinate.
// There is a node for each coordinate (x,y,z), i.e., room.
// The goal is the description of the following scenario:
// - Each node with coordinate z = 0 is a data collector, all other nodes
//   are sensors.
// - All nodes with coordinate (x, y, z!=0) measure temperature and forward
//   packets generated to the node with coordinate (x, y, 0).
// The topology creation is done automatically by the function
//        void CoreExtension_t::createTopology( const TopologySetup_base_t & s )
// see details in CoreTopologySetup_t.hh header file.
//
// The steps of the program are defined below:
// STEP 1: Define topology parameters: How many nodes for each coordinate?
// STEP 2: Define the distance between nodes for each coordinate.
// STEP 3: Define a CoreTopologySetup_t object to store the topology parameters
//         needed for the automatic creation.
// STEP 4: Define the channel setup: in this example only one SHARED channel.
// STEP 5: Call the function createTopology() and store the channel and nodes
//         generated by the call.
// STEP 6: Creation of instances of tasks MyCollector and MySensor.
// STEP 7: Define Communication Protocol: in this example MAC 802.15.4 protocol.
// STEP 8: Task binding: Bind tasks to the SHARED channel.
// STEP 9: Tracing of tasks.

#include <sstream>
#include <systemc>
#include <tlm.h>
#include <exception>

#include <scnsl.hh>
#include "MySensor_t.hh"
#include "MyCollector_t.hh"


using namespace Scnsl;
using namespace Scnsl::Core;
using namespace Scnsl::BuiltinPlugin;
using namespace Scnsl::Setup;
using namespace std;
using namespace Scnsl::Tlm;
using namespace Scnsl::Traffic;


template< typename T >
T *** make_matrix(Scnsl::Core::node_factory_infos_t & node_factory_infos)
{
    T *** ret = new T**[node_factory_infos.nodeNumberX];
	for(unsigned int x = 0; x < node_factory_infos.nodeNumberX; ++x)
    {
        ret[x] = new T*[node_factory_infos.nodeNumberY];
		for(unsigned int y = 0; y < node_factory_infos.nodeNumberY; ++y)
        {
            ret[x][y] = new T[node_factory_infos.nodeNumberZ];
        }
    }

    return ret;
}



int sc_main( int argc, char * argv[] ) {

  try {

///////////////////////////////////////////////////////////////////////////////
// STEP 1 & 2: Topology parameters
///////////////////////////////////////////////////////////////////////////////
	Scnsl::Core::node_factory_infos_t node_factory_infos;
	node_factory_infos.nodeNumberX = 5;		// Number of floors
	node_factory_infos.nodeNumberY = 3;		// Number of corridors
	node_factory_infos.nodeNumberZ = 10;	// Number of rooms in each corridor
	node_factory_infos.nodeStepX = 5;		// Distance x between nodes
	node_factory_infos.nodeStepY = 10;		// Distance y between nodes
	node_factory_infos.nodeStepZ = 10;		// Distance z between nodes

	if ((node_factory_infos.nodeNumberX==1 && node_factory_infos.nodeNumberY==1
	  && node_factory_infos.nodeNumberZ==1)||(node_factory_infos.nodeNumberX==0
	  || node_factory_infos.nodeNumberY==0 || node_factory_infos.nodeNumberZ==0)
	  || node_factory_infos.nodeNumberZ == 1)
		throw std::domain_error( "Incorrect topology parameter." );

	unsigned int nodes_on_floor = node_factory_infos.nodeNumberY *
									node_factory_infos.nodeNumberZ;
	unsigned int nodes_on_row  = node_factory_infos.nodeNumberZ;

///////////////////////////////////////////////////////////////////////////////
// Singleton.
///////////////////////////////////////////////////////////////////////////////
	Scnsl::Setup::Scnsl_t * scnsl = Scnsl::Setup::Scnsl_t::get_instance();

///////////////////////////////////////////////////////////////////////////////
// STEP 3: Setup Topology type:
///////////////////////////////////////////////////////////////////////////////
	CoreTopologySetup_t cts;
	cts.topology_type = CoreTopologySetup_t::MESH;
	cts.extensionId = "core";
	cts.node_factory_infos = node_factory_infos;

///////////////////////////////////////////////////////////////////////////////
// STEP 4: Setup wireless Channel:
///////////////////////////////////////////////////////////////////////////////
	CoreChannelSetup_t ccs;
	ccs.extensionId = "core";
	ccs.channel_type = CoreChannelSetup_t::SHARED;
	ccs.name = "SharedChannel";
	ccs.alpha = 0.1;
	ccs.delay = sc_core::sc_time( 1.0, sc_core::SC_US );
	ccs.nodes_number = node_factory_infos.nodeNumberX *
			node_factory_infos.nodeNumberY * node_factory_infos.nodeNumberZ;
	cts.channel_list.push_back( & ccs );
    
    Scnsl::Utils::DefaultEnvironment_t::createInstance(ccs.alpha);
///////////////////////////////////////////////////////////////////////////////
// STEP 5: BindSetup & createTopology
///////////////////////////////////////////////////////////////////////////////
	BindSetup_base_t *** bsb = make_matrix<BindSetup_base_t>(node_factory_infos);
	for(unsigned int x = 0; x < node_factory_infos.nodeNumberX; ++x)
		for(unsigned int y = 0; y < node_factory_infos.nodeNumberY; ++y)
			for(unsigned int z = 0; z < node_factory_infos.nodeNumberZ; ++z)
			{
				bsb[x][y][z].extensionId = "core";
				bsb[x][y][z].node_binding.bitrate =
							Scnsl::Protocols::Mac_802_15_4::BITRATE;
				bsb[x][y][z].node_binding.transmission_power = 20;
				bsb[x][y][z].node_binding.receiving_threshold = 10;

				cts.bind_list.push_back( & bsb[x][y][z] );
			}

	scnsl->createTopology(cts);

	std::vector< Node_t * > nodes = scnsl->getNodes();
	std::vector< Channel_if_t * > channel_list = scnsl->getChannels();

///////////////////////////////////////////////////////////////////////////////
// STEP 6: Task & Task id creation
///////////////////////////////////////////////////////////////////////////////
	const Scnsl::Core::size_t ONEPROXY = 1;
	const Scnsl::Core::size_t OTHERSPROXIES = nodes_on_row -1;
	unsigned int k=0;
	Scnsl::Core::task_id_t *** id = make_matrix<Scnsl::Core::task_id_t>(node_factory_infos);
	// Transmit nodes -- Z should be -1
	MySensor_t  * *** sensor = make_matrix<MySensor_t *>(node_factory_infos);
	// Receive nodes
	MyCollector_t * **collector = new MyCollector_t **[node_factory_infos.nodeNumberX];

	for(unsigned int x = 0; x < node_factory_infos.nodeNumberX; ++x)
    {
        collector[x] = new MyCollector_t *[node_factory_infos.nodeNumberY];
		for(unsigned int y = 0; y < node_factory_infos.nodeNumberY; ++y)
			for(unsigned int z = 0; z < node_factory_infos.nodeNumberZ; ++z)
			{
				id[x][y][z] = static_cast<const Scnsl::Core::task_id_t>(k++);

				if( z==0 )
				{
					std::stringstream ss;
					ss<< "Collector_"<<y<<"_"<<x;
					collector[x][y] = new MyCollector_t( ss.str().c_str(),
						id[x][y][z], nodes[x*nodes_on_floor+y*nodes_on_row],
						OTHERSPROXIES );
				}
				else
				{
					std::stringstream ss;
					ss<< "Sensor_"<<y<<"_"<<x<<"_"<<z;
					sensor[x][y][z -1] = new MySensor_t( ss.str().c_str(),
						id[x][y][z], nodes[x*nodes_on_floor+y*nodes_on_row+z],
						ONEPROXY);
				}
			}
    }

///////////////////////////////////////////////////////////////////////////////
// STEP 7: Creating the protocol 802.15.4:
///////////////////////////////////////////////////////////////////////////////
	CoreCommunicatorSetup_t ccoms;
	ccoms.extensionId = "core";
	ccoms.ack_required = true;
	ccoms.short_addresses = true;
	ccoms.type = CoreCommunicatorSetup_t::MAC_802_15_4;

	Scnsl::Core::Communicator_if_t * ***mac =
    make_matrix<Scnsl::Core::Communicator_if_t *>(node_factory_infos);

	for(unsigned int x = 0; x < node_factory_infos.nodeNumberX; ++x)
		for(unsigned int y = 0; y < node_factory_infos.nodeNumberY; ++y)
			for(unsigned int z = 0; z < node_factory_infos.nodeNumberZ; ++z)
			{
				ccoms.node = nodes[x * nodes_on_floor + y * nodes_on_row + z];
				mac[x][y][z] = scnsl->createCommunicator( ccoms );
			}

///////////////////////////////////////////////////////////////////////////////
// STEP 8: Task binding
///////////////////////////////////////////////////////////////////////////////
	for(unsigned int x = 0; x < node_factory_infos.nodeNumberX; ++x)
		for(unsigned int y = 0; y < node_factory_infos.nodeNumberY; ++y)
			for(unsigned int z = 1; z < node_factory_infos.nodeNumberZ; ++z)
			{
				scnsl->bind( collector[x][y], NULL, channel_list[0],
					bsb[x][y][0], mac[x][y][0] );
				scnsl->bind( sensor[x][y][z -1], collector[x][y],
					channel_list[0], bsb[x][y][z], mac[x][y][z] );
			}

///////////////////////////////////////////////////////////////////////////////
// STEP 9: Tracing
///////////////////////////////////////////////////////////////////////////////
	// Adding tracing features:
	CoreTracingSetup_t ct;
	ct.extensionId = "core";
	// - Setting the formatter:
	ct.formatterExtensionId = "core";
	ct.formatterName = "basic";
	// Setting the filter:
	ct.filterExtensionId = "core";
	ct.filterName = "basic";
	// Core formatter specific option:
	// printing also the type of trace:
	ct.print_trace_type = true;
	ct.info = 5;
	ct.debug = 0;
	ct.log = 0;
	ct.error = 5;
	ct.warning = 5;
	ct.fatal = 5;
	// Core formatter specific option:
	//  printing also the type of trace:
	ct.print_trace_type = true;
	// - Creating:
	Scnsl_t::Tracer_t * tracer = scnsl->createTracer( ct );
	tracer->addOutput( & std::cout );

	for(unsigned int x = 0; x < node_factory_infos.nodeNumberX; ++x)
		for(unsigned int y = 0; y < node_factory_infos.nodeNumberY; ++y)
			for(unsigned int z = 0; z < node_factory_infos.nodeNumberZ; ++z)
				if ( z == 0 )
					tracer->trace( collector[x][y]);
				else
					tracer->trace( sensor[x][y][z-1]);


	sc_core::sc_start( sc_core::sc_time( 500, sc_core::SC_MS ) );
	sc_core::sc_stop();
  }
  catch ( std::exception & e) {
	cout << e.what() << endl;
	cout << argc << endl;
	cout << argv[0] << endl;
  }

  return 0;
}
